/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <app_control_internal.h>
#include <dlog.h>

#include <memory>

#include "component_based/common/exception.h"
#include "component_based/app_control/component_based_app_control.h"
#include "component_based/app_control/component_based_app_control_implementation.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "CAPI_APPFW_COMPONENT_BASE"

namespace component_based {

AppControl::Impl::Impl(AppControl* parent, app_control_h control)
  : parent_(parent), handle_(control) {
}

AppControl::Impl::~Impl() {
  if (handle_ != nullptr)
    app_control_destroy(handle_);
}

AppControl::Builder::~Builder() = default;

AppControl::Builder& AppControl::Builder::SetAppID(std::string appid) {
  appid_ = std::move(appid);
  return *this;
}

AppControl::Builder& AppControl::Builder::SetOperation(std::string op) {
  op_ = std::move(op);
  return *this;
}

AppControl::Builder& AppControl::Builder::SetUri(Uri uri) {
  uri_ = std::move(uri);
  return *this;
}

AppControl::Builder& AppControl::Builder::SetMime(std::string mime) {
  mime_ = std::move(mime);
  return *this;
}

AppControl::Builder& AppControl::Builder::AddExtra(std::string key,
                                                   std::string value) {
  extra_[std::move(key)] = std::move(value);
  return *this;
}

AppControl::Builder& AppControl::Builder::SetLaunchMode(
    AppControl::LaunchMode mode) {
  mode_ = std::move(mode);
  return *this;
}

AppControl::Builder& AppControl::Builder::SetComponentID(std::string comp_id) {
  comp_id_ = std::move(comp_id);
  return *this;
}

app_control_h AppControl::Builder::BuildNative() {
  app_control_h handle = nullptr;

  app_control_create(&handle);
  if (handle == nullptr)
    return nullptr;

  std::unique_ptr<std::remove_pointer<app_control_h>::type,
      decltype(app_control_destroy)*> ptr(handle, app_control_destroy);
  if (!appid_.empty()) {
    if (app_control_set_app_id(handle, appid_.c_str()) !=
        APP_CONTROL_ERROR_NONE)
      return nullptr;
  }

  if (!op_.empty()) {
    if (app_control_set_operation(handle, op_.c_str()) !=
        APP_CONTROL_ERROR_NONE)
      return nullptr;
  }

  if (!mime_.empty()) {
    if (app_control_set_mime(handle, mime_.c_str()) !=
        APP_CONTROL_ERROR_NONE)
      return nullptr;
  }

  if (app_control_set_uri(handle, uri_.Encode().c_str()) !=
      APP_CONTROL_ERROR_NONE)
      return nullptr;

  for (auto& data : extra_) {
    if (app_control_add_extra_data(handle, data.first.c_str(),
        data.second.c_str()) !=
        APP_CONTROL_ERROR_NONE)
      return nullptr;
  }

  if (app_control_set_launch_mode(handle,
    static_cast<app_control_launch_mode_e>(mode_)) !=
      APP_CONTROL_ERROR_NONE)
      return nullptr;

  if (!comp_id_.empty()) {
    if (app_control_set_component_id(handle, comp_id_.c_str()) !=
        APP_CONTROL_ERROR_NONE)
      return nullptr;
  }
  return ptr.release();
}

AppControl AppControl::Builder::Build() {
  app_control_h handle = BuildNative();
  if (handle == nullptr) {
    THROW(Exception::ErrorCodes::IOError);
  }
  return AppControl(handle);
}

AppControl::AppControl(app_control_h control)
  : impl_(new Impl(this, control)) {
}

AppControl::~AppControl() {
}

AppControl::AppControl(const AppControl& control) {
  app_control_h handle = nullptr;
  if (app_control_clone(&handle, control.impl_->handle_) !=
      APP_CONTROL_ERROR_NONE) {
    THROW(Exception::ErrorCodes::OutOfMemory);
  }
  impl_ = std::unique_ptr<Impl>(new Impl(this, handle));
}

AppControl& AppControl::operator=(const AppControl& control) {
  if (this != &control) {
    app_control_h handle = nullptr;
    if (app_control_clone(&handle, control.impl_->handle_) !=
        APP_CONTROL_ERROR_NONE) {
      THROW(Exception::ErrorCodes::OutOfMemory);
    }
    impl_->handle_ = handle;
  }
  return *this;
}

AppControl::AppControl(AppControl&& control) {
  impl_ = std::unique_ptr<Impl>(new Impl(this, control.impl_->handle_));
  control.impl_->handle_ = nullptr;
}

AppControl& AppControl::operator=(AppControl&& control) {
  if (this != &control) {
    impl_->handle_ = control.impl_->handle_;
    control.impl_->handle_ = nullptr;
  }
  return *this;
}

std::string AppControl::GetAppID() {
  if (impl_->handle_ == nullptr)
    return std::string("");

  char* appid = nullptr;
  app_control_get_app_id(impl_->handle_, &appid);
  if (appid == nullptr)
    return std::string("");

  auto p = std::unique_ptr<char, decltype(std::free)*>(appid, std::free);
  return std::string(p.get());
}

std::string AppControl::GetOperation() {
  if (impl_->handle_ == nullptr)
    return std::string("");

  char* op = nullptr;
  app_control_get_operation(impl_->handle_, &op);
  if (op == nullptr)
    return std::string("");

  auto p = std::unique_ptr<char, decltype(std::free)*>(op, std::free);
  return std::string(p.get());
}

Uri AppControl::GetUri() {
  if (impl_->handle_ == nullptr)
    return Uri("");

  char* uri = nullptr;
  app_control_get_uri(impl_->handle_, &uri);
  if (uri == nullptr)
    return Uri("");

  auto p = std::unique_ptr<char, decltype(std::free)*>(uri, std::free);
  return Uri(std::string(p.get()));
}

std::string AppControl::GetMime() {
  if (impl_->handle_ == nullptr)
    return std::string("");

  char* mime = nullptr;
  app_control_get_mime(impl_->handle_, &mime);
  if (mime == nullptr)
    return std::string("");

  auto p = std::unique_ptr<char, decltype(std::free)*>(mime, std::free);
  return std::string(p.get());
}

std::string AppControl::GetExtra(std::string key) {
  if (impl_->handle_ == nullptr)
    return std::string("");

  char* value = nullptr;
  app_control_get_extra_data(impl_->handle_, key.c_str(), &value);
  if (value == nullptr)
    return std::string("");

  auto p = std::unique_ptr<char, decltype(std::free)*>(value, std::free);
  return std::string(p.get());
}

AppControl::LaunchMode AppControl::GetLaunchMode() {
  if (impl_->handle_ == nullptr)
    return AppControl::LaunchMode::SINGLE;

  app_control_launch_mode_e launch_mode = APP_CONTROL_LAUNCH_MODE_SINGLE;
  if (app_control_get_launch_mode(impl_->handle_, &launch_mode) !=
      APP_CONTROL_ERROR_NONE) {
    THROW(Exception::ErrorCodes::InvalidParameter);
  }
  return static_cast<AppControl::LaunchMode>(launch_mode);
}

std::string AppControl::GetComponentID() {
  if (impl_->handle_ == nullptr)
    return std::string("");

  char* comp_id = nullptr;
  app_control_get_component_id(impl_->handle_, &comp_id);
  if (comp_id == nullptr)
    return std::string("");

  auto p = std::unique_ptr<char, decltype(std::free)*>(comp_id, std::free);
  return std::string(p.get());
}

app_control_h AppControl::GetHandle() const {
  return impl_->handle_;
}

}  // namespace component_based
