/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <appcore_multiwindow_base.h>
#include <app_control_internal.h>
#include <app_common.h>
#include <aul.h>
#include <aul_rpc_port.h>
#include <dlog.h>
#include <Ecore_Wl2.h>
#include <bundle_cpp.h>

#include <memory>

#include "component_based/common/exception.h"
#include "component_based/base/application_base_implementation.h"
#include "component_based/base/application_base.h"
#include "component_based/base/component_manager_internal.h"
#include "component_based/base/dlog_internal.h"

namespace component_based {

ApplicationBase::Impl::Impl(ApplicationBase* parent)
  : parent_(parent), core_base_(new internal::CoreBase(this)) {
  LOGD("");
  char* id = nullptr;
  int r = app_get_id(&id);
  if (r != APP_ERROR_NONE)
    THROW(Exception::ErrorCodes::IOError);

  auto p = std::unique_ptr<char, decltype(std::free)*>(id, std::free);
  app_id_ = std::string(p.get());
}

void ApplicationBase::Impl::OnBaseInit(int argc, char** argv) {
  LOGD("");
  parent_->OnInit(argc, argv);
}

void ApplicationBase::Impl::OnBaseFinish() {
}

void ApplicationBase::Impl::OnBaseRun() {
  LOGD("");
  parent_->OnRun();
}

void ApplicationBase::Impl::OnBaseExit() {
  LOGD("");
  if (exiting_) {
    LOGD("Already exiting");
    return;
  }

  exiting_ = true;
  parent_->OnExit();
}

bool ApplicationBase::Impl::OnBaseCreate() {
  LOGD("");
  auto facs = parent_->OnCreate();
  if (facs.size() == 0) {
    LOGW("Failed to create component factory");
    return false;
  }

  auto iter = facs.begin();
  while (iter != facs.end()) {
    auto i = iter;
    iter++;

    auto comp_id = i->first;
    auto factory = std::move(i->second);
    auto& mgr = internal::ComponentManager::GetInst();
    try {
      mgr.RegisterComponent(std::move(comp_id), std::move(factory));
    } catch(Exception& ex) {
      LOGE("Exception occurred (%s)", ex.what());
      throw;
    }
  }

  return true;
}

void ApplicationBase::Impl::OnBaseTerminate() {
  LOGD("");
  auto& mgr = internal::ComponentManager::GetInst();
  mgr.ExitAll();
  parent_->OnTerminate();
}

void ApplicationBase::Impl::OnBaseReceive(aul_type type, bundle* b) {
  LOGD("type : %d", type);

  tizen_base::Bundle data;
  if (b)
    data = std::move(tizen_base::Bundle(b));
  std::string comp_id = data.GetString(AUL_K_COMPONENT_ID);
  std::string inst_id = data.GetString(AUL_K_INSTANCE_ID);
  std::string rpc_port;
  auto& mgr = internal::ComponentManager::GetInst();
  LOGD("component(%s), instance(%s)", comp_id.c_str(), inst_id.c_str());
  switch (type) {
  case AUL_START:
    rpc_port = data.GetString(AUL_K_RPC_PORT);
    if (rpc_port.empty())
      mgr.Resume(inst_id);
    mgr.Resume(inst_id);
    mgr.Start(comp_id, inst_id, b);
    break;
  case AUL_RESUME:
    if (inst_id.empty()) {
      mgr.ResumeAll();
    } else {
      mgr.Resume(inst_id);
    }
    break;
  case AUL_PAUSE:
    if (inst_id.empty()) {
      mgr.PauseAll();
    } else {
      mgr.Pause(inst_id);
    }
    break;
  case AUL_TERMINATE_INST:
    mgr.Exit(inst_id);
    if (mgr.IsEmpty())
      Exit();
    break;
  case AUL_TERMINATE_BG_INST:
    if (!mgr.IsResumed(inst_id)) {
      mgr.Exit(inst_id);
      if (mgr.IsEmpty())
        Exit();
    }
    break;
  case AUL_TERMINATE:
    break;
  case AUL_TERMINATE_BGAPP:
    if (!core_base_->IsResumed()) {
      Exit();
    }
    break;
  case AUL_WAKE:
    break;
  case AUL_SUSPEND:
    break;
  case AUL_UPDATE_REQUESTED:
    break;
  default:
    break;
  }
}

void ApplicationBase::Impl::OnBaseControl(bundle* b) {
  LOGD("");
}

void ApplicationBase::Impl::OnWindowShow(int type, void* event) {
  Ecore_Wl2_Event_Window_Show* ev =
      static_cast<Ecore_Wl2_Event_Window_Show*>(event);
  LOGD("win(%u), parent_win(%u), event_win(%u)",
      ev->win, ev->parent_win, ev->event_win);
  if (ev->data[0] != 0) {
    auto& mgr = internal::ComponentManager::GetInst();
    std::string inst_id = mgr.GetInstanceID(ev->win);
    mgr.AddGroup(inst_id, ev->data[0]);
  }
}

void ApplicationBase::Impl::OnWindowHide(int type, void* event) {
  Ecore_Wl2_Event_Window_Hide* ev =
      static_cast<Ecore_Wl2_Event_Window_Hide*>(event);
  LOGD("win(%u), parent_win(%u), event_win(%u)",
      ev->win, ev->parent_win, ev->event_win);
}

void ApplicationBase::Impl::OnWindowLower(int type, void* event) {
  Ecore_Wl2_Event_Window_Lower* ev =
      static_cast<Ecore_Wl2_Event_Window_Lower*>(event);
  LOGD("win(%u), timestamp(%u)", ev->win, ev->timestamp);
  auto& mgr = internal::ComponentManager::GetInst();
  std::string inst_id = mgr.GetInstanceID(ev->win);
  mgr.Stop(inst_id);
  mgr.RemoveGroup(inst_id);
}

void ApplicationBase::Impl::OnWindowVisibility(int type, void* event) {
  Ecore_Wl2_Event_Window_Visibility_Change* ev =
      static_cast<Ecore_Wl2_Event_Window_Visibility_Change*>(event);
  LOGD("win(%u), fully_obscured(%d)", ev->win, ev->fully_obscured);
}

void ApplicationBase::Impl::OnWindowPreVisibility(int type, void* event) {
  Ecore_Wl2_Event_Window_Pre_Visibility_Change* ev =
      static_cast<Ecore_Wl2_Event_Window_Pre_Visibility_Change*>(event);
  LOGD("win(%u), type(%u)", ev->win, ev->type);
}

void ApplicationBase::Impl::OnWindowAuxMessage(int type, void* event) {
  Ecore_Wl2_Event_Aux_Message* ev =
      static_cast<Ecore_Wl2_Event_Aux_Message*>(event);
  LOGD("win(%u), key(%s), val(%s)", ev->win, ev->key, ev->val);
}

int ApplicationBase::Impl::Run(int argc, char** argv) {
  int r = core_base_->Init(argc, argv);
  if (r < 0) {
    LOGE("Failed to initialize application");
    return -1;
  }

  core_base_->Fini();
  parent_->OnFini();

  return 0;
}

void ApplicationBase::Impl::Exit() {
  core_base_->Exit();
}

ApplicationBase::ApplicationBase()
  : impl_(new Impl(this)) {
}

ApplicationBase::~ApplicationBase() {
}

void ApplicationBase::Run(int argc, char** argv) {
  if (impl_->Run(argc, argv) < 0)
    return;
}

std::string ApplicationBase::GetApplicationID() {
  return impl_->GetApplicationID();
}

void ApplicationBase::Exit() {
  impl_->Exit();
}

void ApplicationBase::OnInit(int argc, char** argv) {
}

void ApplicationBase::OnFini() {
}

void ApplicationBase::OnRun() {
}

void ApplicationBase::OnExit() {
}

std::map<std::string, std::unique_ptr<Component::Factory>>
ApplicationBase::OnCreate() {
  std::map<std::string, std::unique_ptr<Component::Factory>> facs;
  return facs;
}

void ApplicationBase::OnTerminate() {
}

}  // namespace component_based
