/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <aul.h>
#include <aul_comp_status.h>
#include <app_control_internal.h>
#include <bundle_cpp.h>

#include <memory>

#include "component_based/common/exception.h"
#include "component_based/base/component.h"
#include "component_based/base/component_implementation.h"
#include "component_based/base/dlog_internal.h"

namespace component_based {

Component::Impl::Impl(Component* parent, Component::Type type,
                      std::string comp_id, std::string inst_id)
  : parent_(parent), type_(std::move(type)),
    comp_id_(std::move(comp_id)), inst_id_(std::move(inst_id)),
    sys_ev_(new internal::SystemEvent(this)) {
  auto& mgr = internal::ComponentManager::GetInst();
  if (mgr.IsSingleInstance(inst_id_))
    content_mgr_.reset(new (std::nothrow) internal::ContentManager(comp_id_));
  else
    content_mgr_.reset(new (std::nothrow) internal::ContentManager(inst_id_));

  mgr.AddEventListener(inst_id_, this);
}

bool Component::Impl::OnCreate() {
  aul_comp_notify_start(inst_id_.c_str());

  if (!parent_->OnBaseCreate())
    return false;

  state_ = State::Created;
  aul_comp_status_update(inst_id_.c_str(), COMP_STATUS_CREATED);

  return true;
}

void Component::Impl::OnDestroy() {
  aul_comp_status_update(inst_id_.c_str(), COMP_STATUS_DESTROYED);

  state_ = State::Dying;
  parent_->OnBaseDestroy();

  aul_comp_notify_exit(inst_id_.c_str());
}

void Component::Impl::OnStart(AppControl control, bool restarted) {
  parent_->OnBaseStart(control, restarted);
  state_ = State::Started;
  aul_comp_status_update(inst_id_.c_str(), COMP_STATUS_STARTED);
}

void Component::Impl::OnResume() {
  parent_->OnBaseResume();
}

void Component::Impl::OnPause() {
  parent_->OnBasePause();
}

void Component::Impl::OnStop() {
  if (state_ == State::None ||
      state_ == State::Created ||
      state_ == State::Stopped)
    return;

  parent_->OnBaseStop();
  state_ = State::Stopped;
}

void Component::Impl::OnAction(std::string action, AppControl app_control) {
  parent_->OnBaseAction(std::move(action), std::move(app_control));
}

void Component::Impl::OnLowMemory(LowMemory::Status status) {
  parent_->OnBaseLowMemory(status);
}

void Component::Impl::OnLowBattery(LowBattery::Status status) {
  parent_->OnBaseLowBattery(status);
}

void Component::Impl::OnDeviceOrientationChanged(
       DeviceOrientation::Orientation orientation) {
  parent_->OnBaseDeviceOrientationChanged(orientation);
}

void Component::Impl::OnLanguageChanged(std::string lang) {
  parent_->OnBaseLanguageChanged(std::move(lang));
}

void Component::Impl::OnRegionFormatChanged(std::string region) {
  parent_->OnBaseRegionFormatChanged(std::move(region));
}

void Component::Impl::OnSuspendedStateChanged(SuspendedState::State state) {
  parent_->OnBaseSuspendedStateChanged(state);
}

bool Component::OnBaseCreate() {
  return true;
}

void Component::OnBaseDestroy() {
}

void Component::OnBaseStart(AppControl control, bool restarted) {
}

void Component::OnBaseResume() {
}

void Component::OnBasePause() {
}

void Component::OnBaseStop() {
}

void Component::OnBaseRestoreContent(tizen_base::Bundle content) {
}

void Component::OnBaseSaveContent(tizen_base::Bundle& content) {
}

void Component::OnBaseAction(std::string action, AppControl app_control) {
}

void Component::OnBaseDeviceOrientationChanged(
    DeviceOrientation::Orientation orientation) {
}

void Component::OnBaseLanguageChanged(std::string language) {
}

void Component::OnBaseRegionFormatChanged(std::string region) {
}

void Component::OnBaseLowBattery(LowBattery::Status status) {
}

void Component::OnBaseLowMemory(LowMemory::Status status) {
}

void Component::OnBaseSuspendedStateChanged(SuspendedState::State state) {
}

void Component::Impl::RegisterAction(std::string app_control_name) {
  auto act_iter = act_map_.find(app_control_name);
  if (act_iter != act_map_.end()) {
    LOGW("Already exists");
    return;
  }

  try {
    std::unique_ptr<internal::Action> action = std::unique_ptr<internal::Action>(
        new (std::nothrow) internal::Action(app_control_name, this));
    if (action.get() == nullptr)
      THROW(Exception::ErrorCodes::OutOfMemory);
    act_map_[app_control_name] = std::move(action);
  } catch(Exception& ex) {
    LOGE("Exception occurred (%s)", ex.what());
    throw;
  }
}

bool Component::Impl::DeregisterAction(std::string app_control_name) {
  auto act_iter = act_map_.find(app_control_name);
  if (act_iter == act_map_.end()) {
    LOGW("Not found");
    return false;
  }

  act_map_.erase(app_control_name);
  return true;
}

void Component::Impl::OnAppControlReply(app_control_h request,
                                        app_control_h reply,
                                        app_control_result_e result,
                                        void* user_data) {
  AppControl::IEventListener* ev =
      static_cast<AppControl::IEventListener*>(user_data);
  app_control_h request_handle = nullptr;
  app_control_clone(&request_handle, request);
  app_control_h reply_handle = nullptr;
  app_control_clone(&reply_handle, reply);
  try {
    ev->OnReplyReceived(AppControl(request_handle), AppControl(reply_handle),
        static_cast<AppControl::Result>(result));
  } catch(Exception& ex) {
    LOGE("Exception occurred (%s)", ex.what());
  }
}

void Component::Impl::OnAppControlResult(app_control_h request,
                                         app_control_error_e result,
                                         void* user_data) {
  AppControl::IEventListener* ev =
      static_cast<AppControl::IEventListener*>(user_data);
  app_control_h request_handle = nullptr;
  app_control_clone(&request_handle, request);
  try {
    ev->OnResultReceived(AppControl(request_handle),
        static_cast<AppControl::Error>(result));
  } catch(Exception& ex) {
    LOGE("Exception occurred (%s)", ex.what());
  }
}

Component::State Component::GetState() {
  return impl_->state_;
}

void Component::SetState(Component::State state) {
  impl_->state_ = state;
}

tizen_base::Bundle Component::GetContent() {
  return impl_->content_mgr_->GetContent();
}

void Component::SetContent(tizen_base::Bundle content) {
  impl_->content_mgr_->SetContent(std::move(content));
}

void Component::Impl::SendAsync(AppControl control,
                                AppControl::IEventListener* ev) {
  if (ev == nullptr)
    THROW(Exception::ErrorCodes::InvalidParameter);

  app_control_h handle = control.GetHandle();
  if (handle == nullptr)
    THROW(Exception::ErrorCodes::InvalidParameter);

  int r = app_control_set_caller_instance_id(handle, inst_id_.c_str());
  if (r != APP_CONTROL_ERROR_NONE) {
    LOGE("Failed to set caller instance(%s)", inst_id_.c_str());
    THROW(Exception::ErrorCodes::InvalidParameter);
  }

  r = app_control_send_launch_request_async(handle,
      OnAppControlResult, OnAppControlReply, ev);
  if (r != APP_CONTROL_ERROR_NONE) {
    LOGE("Failed to send launch request async. err(%d)", r);
    THROW(Exception::ErrorCodes::InvalidParameter);
  }
}

AppControl Component::Impl::Send(AppControl control) {
  app_control_h reply = nullptr;
  app_control_h handle = control.GetHandle();
  if (handle == nullptr)
    THROW(Exception::ErrorCodes::InvalidParameter);

  int r = app_control_set_caller_instance_id(handle, inst_id_.c_str());
  if (r != APP_CONTROL_ERROR_NONE) {
    LOGE("Failed to set caller instance(%s)", inst_id_.c_str());
    THROW(Exception::ErrorCodes::InvalidParameter);
  }

  app_control_result_e result;
  r = app_control_send_launch_request_sync(handle, &reply, &result);
  if (r != APP_CONTROL_ERROR_NONE) {
    LOGE("Failed to send launch request sync. err(%d)", r);
    THROW(Exception::ErrorCodes::IOError);
  }

  try {
    AppControl control(reply);
    return control;
  } catch(Exception& ex) {
    LOGE("Exception occurred (%s)", ex.what());
    throw;
  }
}

Component::Component(Component::Type type, std::string comp_id,
                     std::string inst_id)
  : impl_(new Impl(this, std::move(type), std::move(comp_id),
                   std::move(inst_id))) {
}

Component::~Component() = default;

std::string Component::GetComponentID() {
  return impl_->GetComponentID();
}

std::string Component::GetInstanceID() {
  return impl_->GetInstanceID();
}

void Component::Finish() {
  auto& mgr = internal::ComponentManager::GetInst();
  mgr.ExitAtIdle(impl_->GetInstanceID());
}

Component::Type Component::GetType() {
  return impl_->GetType();
}

void Component::RegisterAction(std::string app_control_name) {
  try {
    impl_->RegisterAction(std::move(app_control_name));
  } catch(Exception& ex) {
    LOGE("Exception occurred (%s)", ex.what());
    throw;
  }
}

bool Component::DeregisterAction(std::string app_control_name) {
  return impl_->DeregisterAction(std::move(app_control_name));
}

void Component::SendAsync(AppControl control, AppControl::IEventListener* ev) {
  try {
    impl_->SendAsync(std::move(control), ev);
  } catch(Exception& ex) {
    LOGE("Exception occurred (%s)", ex.what());
    throw;
  }
}

AppControl Component::Send(AppControl control) {
  try {
    return impl_->Send(std::move(control));
  } catch(Exception& ex) {
    LOGE("Exception occurred (%s)", ex.what());
    throw;
  }
}

}  // namespace component_based
