/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef COMPONENT_BASED_BASE_COMPONENT_IMPLEMENTATION_H_
#define COMPONENT_BASED_BASE_COMPONENT_IMPLEMENTATION_H_

#include <app_control.h>
#include <bundle_cpp.h>

#include <string>
#include <map>
#include <memory>

#include "component_based/base/action_internal.h"
#include "component_based/base/component.h"
#include "component_based/base/component_manager_internal.h"
#include "component_based/base/system_event_internal.h"
#include "component_based/base/content_manager_internal.h"

namespace component_based {

class Component::Impl : internal::ComponentManager::IEventListener,
                        internal::SystemEvent::IEventListener,
                        internal::Action::IEventListener {
 public:

  virtual ~Impl() = default;
  bool OnCreate() override;
  void OnDestroy() override;
  void OnStart(AppControl control, bool restarted) override;
  void OnResume() override;
  void OnPause() override;
  void OnStop() override;

  void OnAction(std::string action, AppControl app_control) override;

  void OnLowMemory(LowMemory::Status status) override;
  void OnLowBattery(LowBattery::Status status) override;
  void OnDeviceOrientationChanged(
      DeviceOrientation::Orientation orientation) override;
  void OnLanguageChanged(std::string lang) override;
  void OnRegionFormatChanged(std::string region) override;
  void OnSuspendedStateChanged(SuspendedState::State state) override;

  void RegisterAction(std::string app_control_name);
  bool DeregisterAction(std::string app_control_name);

  const Component::Type& GetType() {
    return type_;
  }

  const std::string& GetComponentID() {
    return comp_id_;
  }

  const std::string& GetInstanceID() {
    return inst_id_;
  }

  void SendAsync(AppControl control, AppControl::IEventListener* ev);
  AppControl Send(AppControl control);

 private:
  static void OnAppControlReply(app_control_h request, app_control_h reply,
                                app_control_result_e result, void* user_data);
  static void OnAppControlResult(app_control_h request,
                                 app_control_error_e result, void* user_data);

 private:
  friend class Component;
  Impl(Component* parent, Component::Type type,
       std::string comp_id, std::string inst_id);

 private:
  Component* parent_;
  Component::Type type_;
  std::string comp_id_;
  std::string inst_id_;
  std::unique_ptr<internal::SystemEvent> sys_ev_;
  std::unique_ptr<internal::ContentManager> content_mgr_;
  State state_ = State::None;
  std::map<std::string, std::unique_ptr<internal::Action>> act_map_;
};

}  // namespace component_base

#endif  // COMPONENT_BASED_BASE_COMPONENT_IMPLEMENTATION_H_
