/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef COMPONENT_BASED_BASE_CONTENT_MANAGER_INTERNAL_H_
#define COMPONENT_BASED_BASE_CONTENT_MANAGER_INTERNAL_H_

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <app_common.h>
#include <bundle_cpp.h>

#include <memory>
#include <string>
#include <vector>

namespace component_based {
namespace internal {

class ContentManager {
 public:
  explicit ContentManager(std::string name);
  virtual ~ContentManager();

  void SetContent(tizen_base::Bundle content);
  tizen_base::Bundle GetContent();

 private:
  class Path {
   private:
    Path() = default;
    ~Path() = default;

   public:
    static Path& GetInst() {
      static Path inst;
      if (!inst.initialized_)
        inst.Init();
      return inst;
    }

    const std::string& GetPath() {
      return path_;
    }

    std::string GetPath(std::string file_name) {
      return path_ + file_name;
    }

   private:
    int Init() {
      char* data_path = app_get_data_path();
      if (data_path) {
        path_ = std::string(data_path) + ".content/";
        free(data_path);
      } else {
        path_ = std::string("/run/aul/dbspace/.content/");
      }

      if (access(path_.c_str(), F_OK) != 0) {
        int r = mkdir(path_.c_str(),
            (S_IRWXU | S_IRGRP | S_IXGRP | S_ISGID | S_IROTH | S_IXOTH));
        if (r < 0)
          return -1;
      }

      initialized_ = true;
      return 0;
    }

   private:
    bool initialized_ = false;
    std::string path_;
  };

  class File {
   public:
    explicit File(std::string path);
    ~File();

    void SetRawData(std::vector<uint8_t> raw_data);
    const std::vector<uint8_t>& GetRawData();

    int Write();
    int Read();
    int Delete();
    bool IsExisting();

   private:
    int Open(bool readonly = true);
    void Close();
    int Sync();
    bool Lock(bool readonly = true);
    void Unlock();
    int TryWrite();
    int TryRead();
    int TryDelete();

    static const int RETRY_COUNT = 7;
    static const int RETRY_SLEEP_UTIME = 10000;

   private:
    std::string path_;
    std::vector<uint8_t> raw_data_;
    int fd_ = 0;
  };

 private:
  std::unique_ptr<ContentManager::File> file_;
};

}  // namespace internal
}  // namespace component_based

#endif  // COMPONENT_BASED_BASE_CONTENT_MANAGER_INTERNAL_H_
