/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>
#include <glib.h>
#include <unistd.h>
#include <app_control_internal.h>

#include <list>

#include "component_based/base/api/component_common.h"
#include "component_based/common/exception.h"
#include "component_based/base/component.h"
#include "component_based/base/frame_window.h"
#include "component_based/base/frame_component.h"
#include "component_based/base/service_component.h"
#include "component_based/base/application_base.h"
#include "component_based/base/api/component_based_app_base.h"
#include "component_based/base/api/base_component.h"
#include "component_based/base/api/base_frame_component.h"
#include "component_based/base/api/base_service_component.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif
#define LOG_TAG "CAPI_COMPONENT_BASED_APPLICATION"

#ifdef EXPORT_API
#undef EXPORT_API
#endif
#define EXPORT_API __attribute__((visibility("default")))

struct frame_window_s {
  int id;
  void* raw;
};

namespace {

class ComponentClass {
 public:
  void Add(std::string comp_id,
           std::unique_ptr<component_based::Component::Factory> factory) {
    auto iter = facs_.find(comp_id);
    if (iter != facs_.end())
      LOGW("Already exists");

    facs_[comp_id] = std::move(factory);
  }

  std::map<std::string, std::unique_ptr<component_based::Component::Factory>>
  Detach() {
    return std::move(facs_);
  }

 private:
  std::map<std::string, std::unique_ptr<component_based::Component::Factory>> facs_;
};

class StubApplicationBase : public component_based::ApplicationBase {
 private:
  StubApplicationBase() = default;
  ~StubApplicationBase() = default;

 public:
  enum class State {
    LAUNCHED,
    CREATED,
    RUNNING,
    DYING,
  };

  static StubApplicationBase& GetInst() {
    static StubApplicationBase inst;
    return inst;
  }

  void SetLifecycleCallback(component_based_app_base_lifecycle_callback_s cb) {
    cb_ = cb;
  }

  void SetUserData(void* user_data) {
    user_data_ = user_data;
  }

  State GetState() {
    return state_;
  }

 private:
  void OnInit(int argc, char** argv) override {
    component_based::ApplicationBase::OnInit(argc, argv);
    if (cb_.init)
      cb_.init(argc, argv, user_data_);
  }

  void OnFini() override {
    state_ = State::DYING;
    component_based::ApplicationBase::OnFini();
    if (cb_.fini)
      cb_.fini(user_data_);
  }

  void OnRun() override {
    state_ = State::RUNNING;
    component_based::ApplicationBase::OnRun();
    if (cb_.run)
      cb_.run(user_data_);
  }

  void OnExit() override {
    component_based::ApplicationBase::OnExit();
    if (cb_.exit)
      cb_.exit(user_data_);
  }

  std::map<std::string, std::unique_ptr<component_based::Component::Factory>>
  OnCreate() override {
    std::map<std::string, std::unique_ptr<component_based::Component::Factory>> facs;
    component_class_h h = nullptr;
    if (cb_.create)
      h = cb_.create(user_data_);

    if (h == nullptr)
      return facs;

    auto cls = static_cast<ComponentClass*>(h);
    facs = std::move(cls->Detach());
    delete cls;

    state_ = State::CREATED;
    return facs;
  }

  void OnTerminate() override {
    if (cb_.terminate)
      cb_.terminate(user_data_);
  }

 private:
  State state_ = State::LAUNCHED;
  component_based_app_base_lifecycle_callback_s cb_;
  void* user_data_ = nullptr;
};

class StubBaseComponent : public component_based::Component {
 public:
  class Factory : public component_based::Component::Factory {
   public:
    Factory(component_type_e type, base_component_lifecycle_callback_s cb, void* user_data)
      : type_(type), cb_(cb), user_data_(user_data) {
    }

    std::unique_ptr<component_based::Component> Create(std::string comp_id,
        std::string inst_id) override {
      return std::unique_ptr<component_based::Component>(
          new (std::nothrow) StubBaseComponent(type_,
            std::move(comp_id), std::move(inst_id), cb_, user_data_));
    }

   private:
    component_type_e type_;
    base_component_lifecycle_callback_s cb_;
    void* user_data_;
  };

  StubBaseComponent(component_type_e type, std::string comp_id,
                     std::string inst_id,
                     base_component_lifecycle_callback_s cb,
                     void* user_data)
    : component_based::Component((component_based::Component::Type)type,
      std::move(comp_id),
      std::move(inst_id)),
      cb_(cb),
      user_data_(user_data) {
  }

  ~StubBaseComponent() = default;

 public:

  void OnBaseRestoreContent(tizen_base::Bundle content) override {
    if (cb_.restore_content)
      cb_.restore_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseSaveContent(tizen_base::Bundle& content) override {
    if (cb_.save_content)
      cb_.save_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseDeviceOrientationChanged(
      component_based::DeviceOrientation::Orientation orientation) override {
    if (cb_.device_orientation_changed) {
      cb_.device_orientation_changed(this,
          static_cast<component_device_orientation_e>(orientation),
          user_data_);
    }
  }

  void OnBaseLanguageChanged(std::string language) override {
    if (cb_.language_changed)
      cb_.language_changed(this, language.c_str(), user_data_);
  }

  void OnBaseRegionFormatChanged(std::string region) override {
    if (cb_.region_format_changed)
      cb_.region_format_changed(this, region.c_str(), user_data_);
  }

  void OnBaseLowBattery(component_based::LowBattery::Status status) override {
    if (cb_.low_battery) {
      cb_.low_battery(this,
          static_cast<component_low_battery_status_e>(status),
          user_data_);
    }
  }

  void OnBaseLowMemory(component_based::LowMemory::Status status) override {
    if (cb_.low_memory) {
      cb_.low_memory(this,
          static_cast<component_low_memory_status_e>(status),
          user_data_);
    }
  }

  void OnBaseSuspendedStateChanged(
      component_based::SuspendedState::State state) override {
    if (cb_.suspended_state_changed) {
      cb_.suspended_state_changed(this,
          static_cast<component_suspended_state_e>(state),
          user_data_);
    }
  }

 private:
  base_component_lifecycle_callback_s cb_;
  void* user_data_;
};

class StubFrameComponent : public component_based::FrameComponent {
 public:
  class Factory : public component_based::FrameComponent::Factory {
   public:
    Factory(base_frame_component_lifecycle_callback_s cb, void* user_data)
      : cb_(cb), user_data_(user_data) {
    }

    std::unique_ptr<component_based::Component> Create(std::string comp_id,
        std::string inst_id) override {
      return std::unique_ptr<component_based::Component>(
          new (std::nothrow) StubFrameComponent(
            std::move(comp_id), std::move(inst_id), cb_, user_data_));
    }

   private:
    base_frame_component_lifecycle_callback_s cb_;
    void* user_data_;
  };

  StubFrameComponent(std::string comp_id,
                     std::string inst_id,
                     base_frame_component_lifecycle_callback_s cb,
                     void* user_data)
    : component_based::FrameComponent(std::move(comp_id), std::move(inst_id)),
      cb_(cb),
      user_data_(user_data) {
  }

  ~StubFrameComponent() = default;

 public:
  std::unique_ptr<component_based::IWindow> OnCreate() override {
    frame_window_h win = cb_.create(this, user_data_);
    if (win == nullptr) {
      LOGE("window is nullptr");
      return nullptr;
    }

    component_based::FrameWindow* frm_win =
        new (std::nothrow) component_based::FrameWindow(win->id, win->raw);
    if (frm_win == nullptr) {
      LOGE("Out of memory");
      return nullptr;
    }

    return std::unique_ptr<component_based::IWindow>(
        static_cast<component_based::IWindow*>(frm_win));
  }

  void OnStart(component_based::AppControl app_control,
               bool restarted) override {
    if (cb_.start)
      cb_.start(this, app_control.GetHandle(), restarted, user_data_);
  }

  void OnResume() override {
    if (cb_.resume)
      cb_.resume(this, user_data_);
  }

  void OnPause() override {
    if (cb_.pause)
      cb_.pause(this, user_data_);
  }

  void OnStop() override {
    if (cb_.stop)
      cb_.stop(this, user_data_);
  }

  void OnDestroy() override {
    if (cb_.destroy)
      cb_.destroy(this, user_data_);
  }

  void OnBaseRestoreContent(tizen_base::Bundle content) override {
    if (cb_.restore_content)
      cb_.restore_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseSaveContent(tizen_base::Bundle& content) override {
    if (cb_.save_content)
      cb_.save_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseAction(std::string action,
                component_based::AppControl app_control) override {
    if (cb_.action)
      cb_.action(this, action.c_str(), app_control.GetHandle(), user_data_);
  }

  void OnBaseDeviceOrientationChanged(
      component_based::DeviceOrientation::Orientation orientation) override {
    if (cb_.device_orientation_changed) {
      cb_.device_orientation_changed(this,
          static_cast<component_device_orientation_e>(orientation),
          user_data_);
    }
  }

  void OnBaseLanguageChanged(std::string language) override {
    if (cb_.language_changed)
      cb_.language_changed(this, language.c_str(), user_data_);
  }

  void OnBaseRegionFormatChanged(std::string region) override {
    if (cb_.region_format_changed)
      cb_.region_format_changed(this, region.c_str(), user_data_);
  }

  void OnBaseLowBattery(component_based::LowBattery::Status status) override {
    if (cb_.low_battery) {
      cb_.low_battery(this,
          static_cast<component_low_battery_status_e>(status),
          user_data_);
    }
  }

  void OnBaseLowMemory(component_based::LowMemory::Status status) override {
    if (cb_.low_memory) {
      cb_.low_memory(this,
          static_cast<component_low_memory_status_e>(status),
          user_data_);
    }
  }

  void OnBaseSuspendedStateChanged(
      component_based::SuspendedState::State state) override {
    if (cb_.suspended_state_changed) {
      cb_.suspended_state_changed(this,
          static_cast<component_suspended_state_e>(state),
          user_data_);
    }
  }

 private:
  base_frame_component_lifecycle_callback_s cb_;
  void* user_data_;
};

class StubServiceComponent : public component_based::ServiceComponent {
 public:
  class Factory : public component_based::ServiceComponent::Factory {
   public:
    Factory(base_service_component_lifecycle_callback_s cb, void* user_data)
      : cb_(cb), user_data_(user_data) {
    }

    std::unique_ptr<component_based::Component> Create(std::string comp_id,
        std::string inst_id) override {
      return std::unique_ptr<component_based::Component>(
          new (std::nothrow) StubServiceComponent(std::move(comp_id),
              std::move(inst_id), cb_, user_data_));
    }

   private:
    base_service_component_lifecycle_callback_s cb_;
    void* user_data_;
  };

  StubServiceComponent(std::string comp_id,
                       std::string inst_id,
                       base_service_component_lifecycle_callback_s cb,
                       void* user_data)
    : component_based::ServiceComponent(std::move(comp_id), std::move(inst_id)),
      cb_(cb),
      user_data_(user_data) {
  }

  ~StubServiceComponent() = default;

 public:
  bool OnCreate() override {
    return cb_.create(this, user_data_);
  }

  void OnStartCommand(component_based::AppControl app_control,
                      bool restarted) override {
    if (cb_.start_command)
      cb_.start_command(this, app_control.GetHandle(), restarted, user_data_);
  }

  void OnDestroy() override {
    if (cb_.destroy)
      cb_.destroy(this, user_data_);
  }

  void OnBaseRestoreContent(tizen_base::Bundle content) override {
    if (cb_.restore_content)
      cb_.restore_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseSaveContent(tizen_base::Bundle& content) override {
    if (cb_.save_content)
      cb_.save_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseAction(std::string action,
                component_based::AppControl app_control) override {
    if (cb_.action)
      cb_.action(this, action.c_str(), app_control.GetHandle(), user_data_);
  }

  void OnBaseDeviceOrientationChanged(
      component_based::DeviceOrientation::Orientation orientation) override {
    if (cb_.device_orientation_changed) {
      cb_.device_orientation_changed(this,
          static_cast<component_device_orientation_e>(orientation),
          user_data_);
    }
  }

  void OnBaseLanguageChanged(std::string language) override {
    if (cb_.language_changed)
      cb_.language_changed(this, language.c_str(), user_data_);
  }

  void OnBaseRegionFormatChanged(std::string region) override {
    if (cb_.region_format_changed)
      cb_.region_format_changed(this, region.c_str(), user_data_);
  }

  void OnBaseLowBattery(component_based::LowBattery::Status status) override {
    if (cb_.low_battery) {
      cb_.low_battery(this,
          static_cast<component_low_battery_status_e>(status),
          user_data_);
    }
  }

  void OnBaseLowMemory(component_based::LowMemory::Status status) override {
    if (cb_.low_memory) {
      cb_.low_memory(this,
          static_cast<component_low_memory_status_e>(status),
          user_data_);
    }
  }

  void OnBaseSuspendedStateChanged(
      component_based::SuspendedState::State state) override {
    if (cb_.suspended_state_changed) {
      cb_.suspended_state_changed(this,
          static_cast<component_suspended_state_e>(state),
          user_data_);
    }
  }

 private:
  base_service_component_lifecycle_callback_s cb_;
  void* user_data_;
};

}  // namespace

extern "C" EXPORT_API int base_frame_create_window(frame_window_h* handle,
    int id, void* raw) {
  if (handle == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  frame_window_h h = (frame_window_h)calloc(1, sizeof(struct frame_window_s));
  if (h == nullptr) {
    LOGE("Out of memory");
    return COMPONENT_ERROR_OUT_OF_MEMORY;
  }
  h->id = id;
  h->raw = raw;
  *handle = h;

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int base_frame_window_get_id(frame_window_h handle,
    int* id) {
  if (handle == nullptr || id == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }
  *id = handle->id;
  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int base_frame_window_get_raw(frame_window_h handle,
    void** raw) {
  if (handle == nullptr || raw == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }
  *raw = handle->raw;
  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int base_frame_get_display_status(
    component_h context, component_display_status_e* display_status) {
  if (context == nullptr || display_status == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  if (component->GetType() != component_based::Component::Type::Frame) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  ::StubFrameComponent* frame_component =
    dynamic_cast<::StubFrameComponent*>(component);
  if (frame_component == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  auto status = frame_component->GetDisplayStatus();
  if (status == component_based::FrameComponent::DisplayStatus::Unknown) {
    LOGE("Invalid context");
    return COMPONENT_ERROR_INVALID_CONTEXT;
  } else if (status == component_based::FrameComponent::DisplayStatus::On) {
    *display_status = COMPONENT_DISPLAY_STATUS_ON;
  } else if (status == component_based::FrameComponent::DisplayStatus::Off) {
    *display_status = COMPONENT_DISPLAY_STATUS_OFF;
  }

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int component_based_app_base_main(int argc, char** argv,
    component_based_app_base_lifecycle_callback_s *callback, void* user_data) {
  if (argc < 1 || argv == nullptr || !callback || !callback->create) {
    LOGE("Invalid parameter");
    return APP_ERROR_INVALID_PARAMETER;
  }

  auto& inst = ::StubApplicationBase::GetInst();
  if (inst.GetState() == ::StubApplicationBase::State::RUNNING) {
    LOGE("Already running");
    return APP_ERROR_ALREADY_RUNNING;
  }

  inst.SetLifecycleCallback(*callback);
  inst.SetUserData(user_data);

  inst.Run(argc, argv);

  return APP_ERROR_NONE;
}

extern "C" EXPORT_API int component_based_app_base_exit(void) {
  auto& inst = ::StubApplicationBase::GetInst();
  inst.Exit();
  return APP_ERROR_NONE;
}

extern "C" EXPORT_API component_class_h
component_based_app_base_add_frame_component(
    component_class_h comp_class,
    const char* comp_id,
    base_frame_component_lifecycle_callback_s* callback,
    void* user_data) {
  if (comp_id == nullptr || callback == nullptr) {
    LOGE("Invalid parameter");
    set_last_result(APP_ERROR_INVALID_PARAMETER);
    return nullptr;
  }

  if (callback->create == nullptr) {
    LOGE("Invalid parameter");
    set_last_result(APP_ERROR_INVALID_PARAMETER);
    return nullptr;
  }

  ::ComponentClass* handle;
  if (comp_class == nullptr) {
    handle = new (std::nothrow) ::ComponentClass();
    if (handle == nullptr) {
      LOGE("Out of memory");
      set_last_result(APP_ERROR_OUT_OF_MEMORY);
      return nullptr;
    }
  } else {
    handle = static_cast<::ComponentClass*>(comp_class);
  }

  ::StubFrameComponent::Factory* factory =
      new (std::nothrow) ::StubFrameComponent::Factory(*callback, user_data);
  if (factory == nullptr) {
    LOGE("Out of memory");
    set_last_result(APP_ERROR_OUT_OF_MEMORY);
    return nullptr;
  }

  handle->Add(comp_id,
      std::unique_ptr<component_based::Component::Factory>(factory));

  comp_class = static_cast<component_class_h>(handle);
  return comp_class;
}

extern "C" EXPORT_API component_class_h
component_based_app_base_add_service_component(
    component_class_h comp_class,
    const char* comp_id,
    base_service_component_lifecycle_callback_s* callback,
    void* user_data) {
  if (comp_id == nullptr || callback == nullptr) {
    LOGE("Invalid parameter");
    set_last_result(APP_ERROR_INVALID_PARAMETER);
    return nullptr;
  }

  if (callback->create == nullptr) {
    LOGE("Invalid parameter");
    set_last_result(APP_ERROR_INVALID_PARAMETER);
    return nullptr;
  }

  ::ComponentClass* handle;
  if (comp_class == nullptr) {
    handle = new ::ComponentClass();
    if (handle == nullptr) {
      LOGE("Out of memory");
      set_last_result(APP_ERROR_OUT_OF_MEMORY);
      return nullptr;
    }
  } else {
    handle = static_cast<::ComponentClass*>(comp_class);
  }

  ::StubServiceComponent::Factory* factory =
      new (std::nothrow) ::StubServiceComponent::Factory(*callback, user_data);
  if (factory == nullptr) {
    LOGE("Out of memory");
    set_last_result(APP_ERROR_OUT_OF_MEMORY);
    return nullptr;
  }

  handle->Add(comp_id,
      std::unique_ptr<component_based::Component::Factory>(factory));

  comp_class = static_cast<component_class_h>(handle);
  return comp_class;
}

extern "C" EXPORT_API component_class_h
component_based_app_base_add_base_component(
    component_class_h comp_class,
    component_type_e type,
    const char* comp_id,
    base_component_lifecycle_callback_s* callback,
    void* user_data) {
  if (comp_id == nullptr || callback == nullptr) {
    LOGE("Invalid parameter");
    set_last_result(APP_ERROR_INVALID_PARAMETER);
    return nullptr;
  }

  ::ComponentClass* handle;
  if (comp_class == nullptr) {
    handle = new ::ComponentClass();
    if (handle == nullptr) {
      LOGE("Out of memory");
      set_last_result(APP_ERROR_OUT_OF_MEMORY);
      return nullptr;
    }
  } else {
    handle = static_cast<::ComponentClass*>(comp_class);
  }

  ::StubBaseComponent::Factory* factory =
      new (std::nothrow) ::StubBaseComponent::Factory(type, *callback, user_data);
  if (factory == nullptr) {
    LOGE("Out of memory");
    set_last_result(APP_ERROR_OUT_OF_MEMORY);
    return nullptr;
  }

  handle->Add(comp_id,
      std::unique_ptr<component_based::Component::Factory>(factory));

  comp_class = static_cast<component_class_h>(handle);
  return comp_class;
}

extern "C" EXPORT_API int component_get_id(component_h context,
    char** id) {
  if (context == nullptr || id == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  *id = strdup(component->GetComponentID().c_str());
  if (*id == nullptr) {
    LOGE("Out of memory");
    return COMPONENT_ERROR_OUT_OF_MEMORY;
  }

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int component_get_instance_id(component_h context,
    char** instance_id) {
  if (context == nullptr || instance_id == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  *instance_id = strdup(component->GetInstanceID().c_str());
  if (*instance_id == nullptr) {
    LOGE("Out of memory");
    return COMPONENT_ERROR_OUT_OF_MEMORY;
  }

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int component_register_action(component_h context,
    const char* action) {
  if (context == nullptr || action == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  try {
    component->RegisterAction(action);
  } catch (component_based::Exception& ex) {
    LOGE("Exception occurred (%s)", ex.what());
    return COMPONENT_ERROR_INVALID_CONTEXT;
  }

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int component_deregister_action(
    component_h context, const char* action) {
  if (context == nullptr || action == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  if (!component->DeregisterAction(action)) {
    LOGE("Deregister fail");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  return COMPONENT_ERROR_NONE;
}

static int __app_control_error_convert(int error) {
  switch (error) {
  case APP_CONTROL_ERROR_INVALID_PARAMETER:
    return COMPONENT_ERROR_INVALID_PARAMETER;
  case APP_CONTROL_ERROR_OUT_OF_MEMORY:
    return COMPONENT_ERROR_OUT_OF_MEMORY;
  case APP_CONTROL_ERROR_APP_NOT_FOUND:
    return COMPONENT_ERROR_NOT_FOUND;
  case APP_CONTROL_ERROR_LAUNCH_REJECTED:
    return COMPONENT_ERROR_LAUNCH_REJECTED;
  case APP_CONTROL_ERROR_LAUNCH_FAILED:
    return COMPONENT_ERROR_LAUNCH_FAILED;
  case APP_CONTROL_ERROR_TIMED_OUT:
    return COMPONENT_ERROR_TIMED_OUT;
  case APP_CONTROL_ERROR_PERMISSION_DENIED:
    return COMPONENT_ERROR_PERMISSION_DENIED;
  default:
    return COMPONENT_ERROR_INVALID_CONTEXT;
  }
}

static int __app_control_send_launch_request_async(
    const char* inst_id,
    app_control_h app_control,
    app_control_result_cb result_cb,
    app_control_reply_cb reply_cb,
    void* user_data) {
  int r = app_control_set_caller_instance_id(app_control, inst_id);
  if (r != APP_CONTROL_ERROR_NONE) {
    LOGE("Failed to set caller instance ID");
    return __app_control_error_convert(r);
  }

  r = app_control_send_launch_request_async(app_control, result_cb, reply_cb,
      user_data);
  if (r != APP_CONTROL_ERROR_NONE) {
    LOGE("Failed to send launch request async. error(%d)", r);
    return __app_control_error_convert(r);
  }

  return COMPONENT_ERROR_NONE;
}

static int __app_control_send_launch_request_sync(
    const char* inst_id,
    app_control_h app_control,
    app_control_h* reply, app_control_result_e* result) {
  int r = app_control_set_caller_instance_id(app_control, inst_id);
  if (r != APP_CONTROL_ERROR_NONE) {
    LOGE("Failed to set caller instance ID");
    return __app_control_error_convert(r);
  }

  r = app_control_send_launch_request_sync(app_control, reply, result);
  if (r != APP_CONTROL_ERROR_NONE) {
    LOGE("Failed to send launch request sync. error(%d)", r);
    return __app_control_error_convert(r);
  }

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int component_send_launch_request_async(
    component_h context,
    app_control_h app_control,
    app_control_result_cb result_cb,
    app_control_reply_cb reply_cb,
    void* user_data) {
  if (context == nullptr || app_control == nullptr || result_cb == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  int r = __app_control_send_launch_request_async(
      component->GetInstanceID().c_str(), app_control, result_cb, reply_cb,
      user_data);
  if (r != COMPONENT_ERROR_NONE) {
    LOGE("Failed to send launch request async. error(%d)", r);
    return r;
  }

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int component_send_launch_request_sync(
    component_h context,
    app_control_h app_control,
    app_control_h* reply, app_control_result_e* result) {
  if (context == nullptr || app_control == nullptr || reply == nullptr ||
      result == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  int r = __app_control_send_launch_request_sync(
      component->GetInstanceID().c_str(), app_control, reply, result);
  if (r != COMPONENT_ERROR_NONE) {
    LOGE("Failed to send launch request sync. error(%d)", r);
    return r;
  }

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int component_finish(component_h context) {
  if (context == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  component->Finish();

  return COMPONENT_ERROR_NONE;
}