/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef COMPONENT_BASED_BASE_SYSTEM_EVENT_INTERNAL_H_
#define COMPONENT_BASED_BASE_SYSTEM_EVENT_INTERNAL_H_

#include <appcore_multiwindow_base.h>
#include <vconf.h>

#include <string>

#include "component_based/base/low_battery.h"
#include "component_based/base/low_memory.h"
#include "component_based/base/device_orientation.h"
#include "component_based/base/suspended_state.h"

namespace component_based {
namespace internal {

class SystemEvent {
 public:
  class IEventListener {
   public:
    virtual void OnLowMemory(LowMemory::Status status) = 0;
    virtual void OnLowBattery(LowBattery::Status status) = 0;
    virtual void OnLanguageChanged(std::string lang) = 0;
    virtual void OnDeviceOrientationChanged(
        DeviceOrientation::Orientation orientation) = 0;
    virtual void OnRegionFormatChanged(std::string region) = 0;
    virtual void OnSuspendedStateChanged(SuspendedState::State state) = 0;
  };

  explicit SystemEvent(IEventListener* ev) : ev_(ev) {
    low_memory_ = appcore_base_add_event(APPCORE_BASE_EVENT_LOW_MEMORY,
        OnLowMemoryCB, ev_);
    low_battery_ = appcore_base_add_event(APPCORE_BASE_EVENT_LOW_BATTERY,
        OnLowBatteryCB, ev_);
    language_changed_ = appcore_base_add_event(APPCORE_BASE_EVENT_LANG_CHANGE,
        OnLanguageChangedCB, ev_);
    device_orientation_changed_ = appcore_base_add_event(
        APPCORE_BASE_EVENT_DEVICE_ORIENTATION_CHANGED,
        OnDeviceOrientationChangedCB, ev_);
    region_format_changed_ = appcore_base_add_event(
        APPCORE_BASE_EVENT_REGION_CHANGE,
        OnRegionFormatChangedCB, ev_);
    suspended_state_changed_ = appcore_base_add_event(
        APPCORE_BASE_EVENT_SUSPENDED_STATE_CHANGE,
        OnSuspendedStateChangedCB, ev_);
  }

  virtual ~SystemEvent() {
    appcore_base_remove_event(suspended_state_changed_);
    appcore_base_remove_event(region_format_changed_);
    appcore_base_remove_event(device_orientation_changed_);
    appcore_base_remove_event(language_changed_);
    appcore_base_remove_event(low_battery_);
    appcore_base_remove_event(low_memory_);
  }

 private:
  static int OnLowMemoryCB(void* event, void* data) {
    IEventListener* ev = static_cast<IEventListener*>(data);
    LowMemory::Status status;
    int* val = static_cast<int*>(event);
    switch (*val) {
    case VCONFKEY_SYSMAN_LOW_MEMORY_NORMAL:
      status = LowMemory::Status::Normal;
      break;
    case VCONFKEY_SYSMAN_LOW_MEMORY_SOFT_WARNING:
      status = LowMemory::Status::SoftWarning;
      break;
    case VCONFKEY_SYSMAN_LOW_MEMORY_HARD_WARNING:
      status = LowMemory::Status::HardWarning;
      break;
    default:
      status = LowMemory::Status::None;
      break;
    }

    ev->OnLowMemory(status);
    return 0;
  }

  static int OnLowBatteryCB(void* event, void* data) {
    IEventListener* ev = static_cast<IEventListener*>(data);
    LowBattery::Status status;
    int* val = static_cast<int*>(event);
    switch (*val) {
    case VCONFKEY_SYSMAN_BAT_POWER_OFF:
      status = LowBattery::Status::PowerOff;
      break;
    case VCONFKEY_SYSMAN_BAT_CRITICAL_LOW:
      status = LowBattery::Status::CriticalLow;
      break;
    default:
      status = LowBattery::Status::None;
      break;
    }

    ev->OnLowBattery(status);
    return 0;
  }

  static int OnLanguageChangedCB(void* event, void* data) {
    IEventListener* ev = static_cast<IEventListener*>(data);
    std::string lang(static_cast<char*>(event));
    ev->OnLanguageChanged(std::move(lang));
    return 0;
  }

  static int OnDeviceOrientationChangedCB(void* event, void* data) {
    IEventListener* ev = static_cast<IEventListener*>(data);
    DeviceOrientation::Orientation orientation;
    int* val = static_cast<int*>(event);
    switch (*val) {
    case APPCORE_BASE_RM_PORTRAIT_NORMAL:
      orientation = DeviceOrientation::Orientation::Orientation_0;
      break;
    case APPCORE_BASE_RM_PORTRAIT_REVERSE:
      orientation = DeviceOrientation::Orientation::Orientation_180;
      break;
    case APPCORE_BASE_RM_LANDSCAPE_NORMAL:
      orientation = DeviceOrientation::Orientation::Orientation_270;
      break;
    case APPCORE_BASE_RM_LANDSCAPE_REVERSE:
      orientation = DeviceOrientation::Orientation::Orientation_90;
      break;
    default:
      orientation = DeviceOrientation::Orientation::Orientation_0;
      break;
    }

    ev->OnDeviceOrientationChanged(orientation);
    return 0;
  }

  static int OnRegionFormatChangedCB(void* event, void* data) {
    IEventListener* ev = static_cast<IEventListener*>(data);
    std::string region(static_cast<char*>(event));
    ev->OnRegionFormatChanged(std::move(region));
    return 0;
  }

  static int OnSuspendedStateChangedCB(void* event, void* data) {
    IEventListener* ev = static_cast<IEventListener*>(data);
    SuspendedState::State state;
    int* val = static_cast<int*>(event);
    if ((*val) == 0) {
      state = SuspendedState::State::Will_Enter;
    } else {
      state = SuspendedState::State::Did_Exit;
    }

    ev->OnSuspendedStateChanged(state);
    return 0;
  }

 private:
  IEventListener* ev_;
  appcore_base_event_h low_memory_;
  appcore_base_event_h low_battery_;
  appcore_base_event_h language_changed_;
  appcore_base_event_h device_orientation_changed_;
  appcore_base_event_h region_format_changed_;
  appcore_base_event_h suspended_state_changed_;
};

}  // namespace internal
}  // namespace component_based

#endif  // COMPONENT_BASED_BASE_SYSTEM_EVENT_INTERNAL_H_
