// Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
// Use of this source code is governed by a apache 2.0 license that can be
// found in the LICENSE file.

#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <stdlib.h>

#include "component_based/app_control/component_based_app_control.h"
#include "component_based/base/frame_component.h"
#include "component_based/base/service_component.h"
#include "component_based/base/component.h"
#include "component_based/base/component_implementation.h"

#include "unit_tests/mock/mock_app_common.h"
#include "unit_tests/mock/mock_app_control.h"
#include "unit_tests/mock/mock_appcore_multiwindow_base.h"
#include "unit_tests/mock/mock_elementary.h"

namespace {

class ComponentTest : public ::testing::Test {
 protected:
  void SetUp() override {}
  void TearDown() override {}
};

class TestListener : public component_based::AppControl::IEventListener {
 public:
  void OnResultReceived(component_based::AppControl request,
                                component_based::AppControl::Error error) {
  }
  void OnReplyReceived(component_based::AppControl request,
      component_based::AppControl reply,
      component_based::AppControl::Result result) {
  }
};

int __fake_app_get_display_state(app_display_state_e* state) {
  *state = APP_DISPLAY_STATE_ON;
  return 0;
}

int __fake_app_resource_manager_get(app_resource_e type, const char* id,
    char** path) {
  char* temp = strdup("tmp");
  *path = temp;
  return 0;
}

Evas_Object* __fake_elm_win_util_standard_add(const char* name, const char* title)
{
  return nullptr;
}

int __fake_app_control_add_action_handler(const char* action,
    app_control_action_cb callback, void *user_data, app_control_action_h* h)
{
  return 0;
}

TEST_F(ComponentTest, TestFrameComponentClass) {
  elm_win_util_standard_add_fake.custom_fake = __fake_elm_win_util_standard_add;
  app_get_display_state_fake.custom_fake = __fake_app_get_display_state;

  std::unique_ptr<component_based::FrameComponent> fc(
      new component_based::FrameComponent(std::string("comp_id"), std::string("inst_id")));

  std::unique_ptr<component_based::IWindow> win = fc->OnCreate();

  std::unique_ptr<component_based::AppControl::Builder> builder(
      new component_based::AppControl::Builder());
  EXPECT_NE(builder, nullptr);

  builder->SetAppID("org.tizen.helloworld");
  builder->SetOperation(APP_CONTROL_OPERATION_DEFAULT);
  builder->SetLaunchMode(component_based::AppControl::LaunchMode::SINGLE);
  builder->SetUri(component_based::Uri("https:%2F%2Fjohn.doe%40www.example.com%3A123%2Fforum%2Fquestions%2F?tag=networking&order=newest#top"));
  builder->SetMime("application/octet-stream");
  builder->AddExtra("key", "value");
  builder->SetComponentID("ElmFrame");

  component_based::AppControl ctrl = builder->Build();
  fc->OnStart(ctrl, false);
  fc->OnResume();
  fc->OnPause();
  fc->OnStop();
  fc->OnDestroy();

  component_based::FrameComponent::DisplayStatus status =
      fc->GetDisplayStatus();
  EXPECT_EQ(status, component_based::FrameComponent::DisplayStatus::On);
  fc->GetWindow();
}

TEST_F(ComponentTest, TestSerivceComponentClass) {
  std::unique_ptr<component_based::ServiceComponent> sc(
      new component_based::ServiceComponent(std::string("comp_id"), std::string("inst_id")));

  sc->OnCreate();

  std::unique_ptr<component_based::AppControl::Builder> builder(
      new component_based::AppControl::Builder());
  EXPECT_NE(builder, nullptr);

  builder->SetAppID("org.tizen.helloworld");
  builder->SetOperation(APP_CONTROL_OPERATION_DEFAULT);
  builder->SetLaunchMode(component_based::AppControl::LaunchMode::SINGLE);
  builder->SetUri(component_based::Uri("https:%2F%2Fjohn.doe%40www.example.com%3A123%2Fforum%2Fquestions%2F?tag=networking&order=newest#top"));
  builder->SetMime("application/octet-stream");
  builder->AddExtra("key", "value");
  builder->SetComponentID("ElmFrame");

  component_based::AppControl ctrl = builder->Build();
  sc->OnStartCommand(ctrl, false);
  sc->OnDestroy();
}

TEST_F(ComponentTest, TestComponentClass) {
  app_resource_manager_get_fake.custom_fake = __fake_app_resource_manager_get;
  app_control_add_action_handler_fake.custom_fake =
    __fake_app_control_add_action_handler;

  std::unique_ptr<component_based::Component> fc(
      new component_based::Component(component_based::Component::Type::Frame, std::string("comp_id"),
      std::string("inst_id")));

  std::string cid = fc->GetComponentID();
  EXPECT_EQ(cid, "comp_id");

  std::string iid = fc->GetInstanceID();
  EXPECT_EQ(iid, "inst_id");

  component_based::Component::Type type = fc->GetType();
  EXPECT_EQ(type, component_based::Component::Type::Frame);
  fc->Finish();

  fc->RegisterAction(std::string("action"));
  fc->DeregisterAction(std::string("action"));

  std::unique_ptr<component_based::AppControl::Builder> builder(
      new component_based::AppControl::Builder());
  EXPECT_NE(builder, nullptr);

  builder->SetAppID("org.tizen.helloworld");
  builder->SetOperation(APP_CONTROL_OPERATION_DEFAULT);
  builder->SetLaunchMode(component_based::AppControl::LaunchMode::SINGLE);
  builder->SetUri(component_based::Uri("https:%2F%2Fjohn.doe%40www.example.com%3A123%2Fforum%2Fquestions%2F?tag=networking&order=newest#top"));
  builder->SetMime("application/octet-stream");
  builder->AddExtra("key", "value");
  builder->SetComponentID("ElmFrame");

  component_based::AppControl ctrl = builder->Build();
  std::unique_ptr<TestListener> listener =
      std::unique_ptr<TestListener>(new TestListener());
  fc->SendAsync(ctrl, listener.get());

  fc->Send(ctrl);
}

}  // namespace
